@@ -1,44 +1,21 @@
 from __future__ import absolute_import
 
-import xapian
 import string
+import xapian
 
-from cStringIO import StringIO
-from cPickle import Pickler, Unpickler
-from xapian import QueryParser, Query, UnimplementedError, DocNotFoundError
+from operator import itemgetter
+from functools import partial
+from json import dumps, loads
+
+from xapian import Query, QueryParser, UnimplementedError, DocNotFoundError
 
 from . import snowball
-from .spies import TermCountMatchSpy
+from .exc import ValidationError
+from .util import OrderedDict
 
 
 RETRY_LIMIT = 3
 
-try:
-    # try bz2, fallback to zlib
-    from bz2 import compress, decompress
-
-    def _decompress(data):
-        if data.startswith('BZ'):
-            try:
-                data = decompress(data)
-            except IOError:
-                pass
-        return data
-
-except ImportError:
-    from zlib import compress, decompress, error as compress_error
-
-    def _decompress(data):
-        if data.startswith('x'):
-            try:
-                data = decompress(data)
-            except compress_error:
-                pass
-        return data
-
-
-from . schema import Schema
-
 
 default_parser_flags = (QueryParser.FLAG_PHRASE |
                         QueryParser.FLAG_BOOLEAN |
@@ -48,7 +25,41 @@ default_parser_flags = (QueryParser.FLAG_PHRASE |
                         QueryParser.FLAG_WILDCARD)
 
 
-class LangDecider(xapian.ExpandDecider):
+def _schema_name(schema):
+    return "%s.%s" % (schema.__module__, schema.__name__)
+
+
+def defaults(head, limit=0, mlimit=0, klimit=1.0, kmlimit=1.0):
+    return (head, limit, mlimit, klimit, kmlimit)
+
+
+def _lookup_schema(name):
+    modname, expr = name.rsplit('.', 1)
+    local_name = modname.split('.')[-1]
+    mod = __import__(modname, {}, {}, local_name)
+    return eval(expr, mod.__dict__)
+
+
+class Record(object):
+
+    def __init__(self, schema):
+        self._xodb_schema = schema
+
+    def __getattr__(self, name):
+        try:
+            return self._xodb_schema[name].value
+        except KeyError:
+            raise AttributeError(name)
+
+    def __repr__(self):
+        return repr(self._xodb_schema)
+
+
+def record_factory(database, record):
+    return Record(record)
+
+
+class LanguageDecider(xapian.ExpandDecider):
     """
     A Xapian ExpandDecider that decide which terms to keep and which
     to discard when expanding a query using the "suggest" syntax.  As
@@ -60,17 +71,17 @@ class LangDecider(xapian.ExpandDecider):
       - Terms shorter than min_length chars, which are likely irrelevant
 
       - Stopwords for the given language.  Default is english, pass
-        None for the lang argument if no stopping is desired.
+        None for the language argument if no stopping is desired.
     """
 
     min_length = 5
     nostart = unicode(string.uppercase + string.digits)
 
-    def __init__(self, lang="en", filter=None, stems=None):
-        super(LangDecider, self).__init__()
-        if lang in snowball.stoppers:
-            self.stopper = snowball.stoppers[lang]
-            self.stemmer = xapian.Stem(lang)
+    def __init__(self, language="en", filter=None, stems=None):
+        super(LanguageDecider, self).__init__()
+        if language in snowball.stoppers:
+            self.stopper = snowball.stoppers[language]
+            self.stemmer = xapian.Stem(language)
         else:
             self.stopper = lambda(term): False
             self.stemmer = xapian.Stem("none")
@@ -93,10 +104,10 @@ class PrefixDecider(xapian.ExpandDecider):
 
     def __init__(self, prefix):
         super(PrefixDecider, self).__init__()
-        self.prefix = 'X' + prefix.upper()
+        self.prefix = (u'X%s:' % prefix.upper()).encode('utf-8')
 
     def __call__(self, term):
-        return term.decode('utf-8').startswith(self.prefix)
+        return term.startswith(self.prefix)
 
 
 class MultipleValueRangeProcessor(xapian.ValueRangeProcessor):
@@ -124,26 +135,8 @@ class MultipleValueRangeProcessor(xapian.ValueRangeProcessor):
         return (xapian.BAD_VALUENO, begin, end)
 
 
-class Record(object):
-    """Default dumb record type."""
-
-    def __init__(self, otype, newargs, state):
-        self.__dict__ = state
-
-
-class BrokenRecord(Record):
-    """Type returned when pickled type cannot be found."""
-    pass
-
-
-def instance_factory(otype, newargs, state):
-    inst = otype.__new__(otype, *newargs)
-    inst.__dict__ = state
-    return inst
-
-
 class Database(object):
-    """A xodb database.
+    """An xodb database.
 
     :param db_or_path: A path to file, or a xapian.Database object
     that backs this Database instance.
@@ -155,35 +148,22 @@ class Database(object):
 
     :param spelling: If True, write spelling correction data to the
     database.
-
-    :param record_factory: Callable factory to pass stored object
-    state when results are instanciated.
-
-    :param persistent_id: Callable to resolve persistent ids for
-    pickled object state.
-
-    :param persistent_load: Callable to resolve pickled object state
-    for persistent id.
     """
-    COMPRESS_OVER = 256 # set to 0 to disable compression
+
+    record_factory = record_factory
 
     relevance_prefix = "_XODB_RP_"
     boolean_prefix = "_XODB_BP_"
     value_prefix = "_XODB_VALUE_"
     value_sort_prefix = "_XODB_VALUESORT_"
     value_count_name = "_XODB_COUNT_"
-    weak_doc_map = None
+    db_path = None
 
     def __init__(self, db_or_path,
-                 writable=True, overwrite=False,
-                 spelling=True, record_factory=Record, weak_doc_map=True,
-                 persistent_id=None, persistent_load=None):
+                 writable=True, overwrite=False, spelling=True):
         self._writable = writable
         self.backend = db_or_path
         self.spelling = spelling
-        self.record_factory = record_factory
-        self.persistent_id = persistent_id
-        self.persistent_load = persistent_load
 
         if isinstance(db_or_path, basestring):
             if writable:
@@ -194,6 +174,7 @@ class Database(object):
                 self.backend = xapian.WritableDatabase(db_or_path, flags)
             else:
                 self.backend = xapian.Database(db_or_path)
+            self.db_path = db_or_path
         elif not isinstance(db_or_path, xapian.Database):
             raise TypeError(
                 'First argument must be path or xapian.[Writable]Database')
@@ -203,56 +184,19 @@ class Database(object):
         self.values = {}
         self.value_sorts = {}
         self.type_map = {}
-
-        self.map(object, Schema)
         self.reopen()
 
     def map(self, otype, schema):
         """Map a type to a schema."""
         self.type_map[otype] = schema
 
-    def unmap(self, otype):
-        """Unmap a type."""
-        del self.type_map[otype]
-
-    def clear_map(self):
-        self.type_map = {}
-
     def schema_for(self, otype):
         """Get the schema for a given type, or one of its
         superclasses."""
         for base in otype.__mro__:
             if base in self.type_map:
                 return self.type_map[base]
-        return Schema
-
-    def serialize(self, typ, newargs, state):
-        f = StringIO()
-        pickler = Pickler(f, -1)
-        pickler.dump((typ, newargs))
-        if self.persistent_id is not None:
-            pickler.persistent_id = self.persistent_id
-        pickler.dump(state)
-        data = f.getvalue()
-        if self.COMPRESS_OVER and len(data) > self.COMPRESS_OVER:
-            data = compress(data)
-        return data
-
-    def unserialize(self, data, record_factory=None):
-        data = _decompress(data)
-        file = StringIO(data)
-        unpickler = Unpickler(file)
-        try:
-            otype, newargs = unpickler.load()
-        except AttributeError:
-            otype, newargs = Record, ()
-        if self.persistent_load is not None:
-            unpickler.persistent_load = self.persistent_load
-        state = unpickler.load()
-        if record_factory:
-            return record_factory(otype, newargs, state)
-        else:
-            return otype, newargs, state
+        raise TypeError("No schema defined for %s" % repr(otype))
 
     def _get_value_count(self):
         return int(self.backend.get_metadata(self.value_count_name))
@@ -263,14 +207,20 @@ class Database(object):
     value_count = property(_get_value_count, _set_value_count)
 
     def add_prefix(self, key, value):
+        """Add a prefix mapping to the database.
+        """
         self.relevance_prefixes[key] = value
         self.backend.set_metadata(self.relevance_prefix + key, value)
 
     def add_boolean_prefix(self, key, value):
+        """Add a boolean prefix mapping to the database.
+        """
         self.boolean_prefixes[key] = value
         self.backend.set_metadata(self.boolean_prefix + key, value)
 
     def add_value(self, name, sort=None):
+        """Add a value mapping to the database.
+        """
         if name in self.values:
             return self.values[name]
         value_count = self.value_count + 1
@@ -288,6 +238,7 @@ class Database(object):
         return self.backend.get_doccount()
 
     def allterms(self, prefix=""):
+        self.backend.reopen()
         for t in self.backend.allterms(prefix):
             yield t.term
 
@@ -301,8 +252,7 @@ class Database(object):
             return default
 
     def __getitem__(self, docid):
-        return self.unserialize(self.backend.get_document(docid).get_data(),
-                                self.record_factory)
+        return self.backend.get_document(docid)
 
     def __delitem__(self, docid):
         self.backend.delete_document(docid)
@@ -314,24 +264,66 @@ class Database(object):
         else:
             self.backend.replace_document(docid, document)
 
-    def add(self, *objs):
+    def __contains__(self, docid):
+        return True if self.get(docid) else False
+
+    def add(self, *objs, **kw):
+        """Add an object to the database by transforming it into a
+        xapian document.  It's type or one of its base types must be
+        mapped to a schema before an object can be added here.
+
+        :params objs: One or more mapped objects to add.
+
+        :param validate: Validated the schema before the object is
+        added.  Default: True
+
+        Returns a list of xapan documents that were added to the
+        database.
+        """
         added = []
+        validate = kw.pop('validate', True)
         for obj in objs:
+            if isinstance(obj, xapian.Document):
+                self.backend.add_document(obj)
+                added.append(obj)
+                continue
+
             if hasattr(obj, '__xodb_schema__'):
                 schema_type = obj.__xodb_schema__
             else:
                 schema_type = self.schema_for(type(obj))
-            schema = schema_type(self)
-            added.append(self.backend.add_document(schema(obj)))
+            schema = schema_type.from_defaults()
+            schema.update_by_object(obj)
+
+            if validate and not schema.validate():
+                invalid = []
+                for child in schema.all_children:
+                    if not child.valid:
+                        invalid.append(child)
+                from pdb import set_trace; set_trace()
+
+                raise ValidationError("Elements of %s did not validate %s:" %
+                                      (schema.__class__.__name__,
+                                       list((c.name, c.value)
+                                            for c in invalid)))
+
+            doc = schema.to_document(self)
+            data = dumps((_schema_name(schema_type), schema.flatten()))
+            doc.set_data(data)
+            self.backend.add_document(doc)
+            added.append(doc)
         return added
 
     def reopen(self):
+        """Reopen the database, refresing the local connection's view
+        of the prefix and value data.
+        """
         self.backend.reopen()
         try:
             keys = self.backend.metadata_keys()
         except UnimplementedError:
-            # Inmemory backends don't expose this function, so just
-            # return since they can't persist anything anyway!
+            # Inmemory backends don't expose metadata, so just return
+            # since they can't persist anything anyway!
             return
 
         self.relevance_prefixes.clear()
@@ -353,8 +345,8 @@ class Database(object):
                 value = k[len(self.value_sort_prefix):]
                 self.value_sorts[value] = self.backend.get_metadata(k)
 
-        # hit the property to refresh this value
         try:
+            # hit the property to refresh this value
             count = self.value_count
         except ValueError:
             if self._writable:
@@ -373,8 +365,9 @@ class Database(object):
         if self._writable:
             self.backend.commit_transaction()
 
-    def prepare_query_parser(self, lang=None, default_op=xapian.Query.OP_AND):
-        qp = xapian.QueryParser()
+    def prepare_query_parser(self, language=None,
+                             default_op=Query.OP_AND):
+        qp = QueryParser()
         qp.set_database(self.backend)
         qp.set_default_op(default_op)
 
@@ -388,49 +381,155 @@ class Database(object):
             # First add numeric values ranges
             qp.add_valuerangeprocessor(MultipleValueRangeProcessor(
                 dict(((k, self.values[k])
-                      for k, v in self.value_sorts.items() if v == 'numeric')),
+                      for k, v in self.value_sorts.items() if v == 'integer')),
                 lambda s: xapian.sortable_serialise(float(s)),
             ))
             # Then string and date
             qp.add_valuerangeprocessor(MultipleValueRangeProcessor(
                 dict(((k, self.values[k])
-                      for k, v in self.value_sorts.items() if v != 'numeric')),
+                      for k, v in self.value_sorts.items() if v != 'integer')),
             ))
-        if lang in snowball.stoppers:
-            qp.set_stemmer(xapian.Stem(lang))
-            qp.set_stopper(snowball.stoppers[lang])
-            qp.set_stemming_strategy(xapian.QueryParser.STEM_SOME)
+        if language in snowball.stoppers:
+            qp.set_stemmer(xapian.Stem(language))
+            qp.set_stopper(snowball.stoppers[language])
+            qp.set_stemming_strategy(QueryParser.STEM_SOME)
         return qp
 
-    def query(self, query, offset=0, limit=None, order=None,
-              reverse=False, lang=None, check=0,
-              match_decider=None, match_spy=None,
+    def querify(self, query, language=None, translit='translit/long',
+                default_op=Query.OP_AND, parser_flags=default_parser_flags):
+        if isinstance(query, Query):
+            return query
+        if isinstance(query, basestring):
+            if query == "":
+                return Query("")
+            else:
+                query = query.lower()
+                if translit:
+                    query = query.encode(translit)
+                qp = self.prepare_query_parser(language, default_op)
+                return qp.parse_query(query, parser_flags)
+        else:
+            return reduce(partial(Query, default_op),
+                          (self.querify(q,
+                                        language,
+                                        translit,
+                                        default_op,
+                                        parser_flags) for q in query))
+
+    def query(self, query, offset=0, limit=0, order=None,
+              reverse=False, language=None, check=0, translit='translit/long',
+              match_decider=None, match_spy=None, document=False,
               echo=False, parser_flags=default_parser_flags,
-              document=False,
-              default_op=xapian.Query.OP_AND):
+              default_op=Query.OP_AND):
         """
         Query the database with the provided string or xapian Query
         object.  A string is passed into xapians QueryParser first to
         generate a Query object.
         """
         self.reopen()
-        if query == "":
-            query = xapian.Query("")
-        else:
-            qp = self.prepare_query_parser(lang, default_op)
-            query = qp.parse_query(query, parser_flags)
 
+        enq = xapian.Enquire(self.backend)
+        query = self.querify(query, language, translit,
+                             default_op, parser_flags)
         if echo:
             print str(query)
-        enq = xapian.Enquire(self.backend)
         enq.set_query(query)
 
+        limit = limit or self.backend.get_doccount()
+
         mset = self._build_mset(enq, offset, limit, order, reverse,
                                 check, match_decider, match_spy)
         return self._return_objects(mset, document=document)
 
-    def estimate(self, query, limit=0, lang=None,
-                 parser_flags=default_parser_flags):
+    def count(self, query="", language=None, echo=False,
+              translit='translit/long',
+              parser_flags=default_parser_flags,
+              default_op=Query.OP_AND):
+        """
+        Query the database with the provided string or xapian Query
+        object.  A string is passed into xapians QueryParser first to
+        generate a Query object.
+        """
+        self.reopen()
+        query = self.querify(query, language, translit, default_op, parser_flags)
+        if echo:
+            print str(query)
+        enq = xapian.Enquire(self.backend)
+        enq.set_query(query)
+
+        mset = self._build_mset(enq)
+        return mset.size()
+
+    def facet(self, query,
+              prefix='facet',
+              estimate=True,
+              language=None,
+              limit=0,
+              mlimit=0,
+              klimit=1.0,
+              kmlimit=1.0,
+              echo=False):
+        """Get facet suggestions for the query, then the query with
+        each suggested facet, asking xapian for an estimated count of
+        each sub-query.
+        """
+        if estimate:
+            counter = self.estimate
+        else:
+            counter = self.count
+
+        results = {}
+        suggestions = self.suggest(query,
+                                   prefix=prefix,
+                                   language=language,
+                                   limit=limit,
+                                   mlimit=mlimit,
+                                   klimit=klimit,
+                                   kmlimit=kmlimit,
+                                   echo=echo)
+        for facet in suggestions:
+            q = self.querify([query, facet])
+            if echo:
+                print str(q)
+            else:
+                results[facet] = counter(q, language=language)
+        return results
+
+    def expand(self, query, expand,
+               language=None,
+               echo=False,
+               translit='translit/long',
+               default_op=Query.OP_AND,
+               parser_flags=default_parser_flags):
+
+        q = self.querify(query, language, translit, default_op, parser_flags)
+        results = {}
+        head, tail = expand[0], expand[1:]
+        if isinstance(head, tuple):
+            args = head
+        else:
+            args = (head,)
+        head, limit, mlimit, klimit, kmlimit = defaults(*args)
+
+        r = self.facet(
+            q, prefix=head, language=language, echo=echo,
+            limit=limit, mlimit=mlimit,
+            klimit=klimit, kmlimit=kmlimit).items()
+
+        for name, score in r:
+            if tail:
+                subq = self.querify([q, name])
+                r = self.expand(
+                    subq, tail, language, echo,
+                    default_op, parser_flags)
+                results[(name, score)] = r
+            else:
+                results[(name, score)] = score
+        return results
+
+    def estimate(self, query, limit=0, klimit=1.0, language=None,
+                 translit='translit/long',
+                 default_op=Query.OP_AND, parser_flags=default_parser_flags):
         """Estimate the number of documents that will be yielded with the
         given query.
 
@@ -440,13 +539,9 @@ class Database(object):
         enq = xapian.Enquire(self.backend)
 
         if limit == 0:
-            limit = self.backend.get_doccount()
+            limit = int(self.backend.get_doccount() * klimit)
 
-        if query == "":
-            query = xapian.Query("")
-        else:
-            qp = self.prepare_query_parser(lang)
-            query = qp.parse_query(query, parser_flags)
+        query = self.querify(query, language, translit, default_op, parser_flags)
 
         enq.set_query(query)
         return enq.get_mset(0, 0, limit).get_matches_estimated()
@@ -459,34 +554,29 @@ class Database(object):
         self.backend.reopen()
         return self.backend.get_termfreq(term)
 
-    def get_doccount(self):
-        """
-        Return the number of indexed documents, handy for tests and
-        sanity check.
-        """
-        self.backend.reopen()
-        return self.backend.get_doccount()
-
-    def describe_query(self, query, lang=None,
-              default_op=xapian.Query.OP_AND):
+    def describe_query(self, query, language=None,
+              default_op=Query.OP_AND):
         """
         Describe the parsed query.
         """
-        qp = self.prepare_query_parser(lang, default_op)
+        qp = self.prepare_query_parser(language, default_op)
         q = qp.parse_query(query, default_parser_flags)
         return str(q)
 
-    def spell(self, query, lang=None):
+    def spell(self, query, language=None):
         """
         Suggest a query string with corrected spelling.
         """
         self.backend.reopen()
-        qp = self.prepare_query_parser(lang)
-        qp.parse_query(query, xapian.QueryParser.FLAG_SPELLING_CORRECTION)
+        qp = self.prepare_query_parser(language)
+        qp.parse_query(query, QueryParser.FLAG_SPELLING_CORRECTION)
         return qp.get_corrected_query_string().decode('utf8')
 
-    def suggest(self, query, offset=0, limit=10, moffset=0, mlimit=10,
-                lang=None, prefix=None, decider=None, score=False):
+    def suggest(self, query, offset=0, limit=0, moffset=0, mlimit=0,
+                klimit=1.0, kmlimit=1.0, translit='translit/long',
+                language=None, prefix=None, decider=None, score=False,
+                echo=False, default_op=Query.OP_AND,
+                parser_flags=default_parser_flags):
         """
         Suggest terms that would possibly yield more relevant results
         for the given query.
@@ -494,12 +584,14 @@ class Database(object):
         self.reopen()
         enq = xapian.Enquire(self.backend)
 
-        qp = self.prepare_query_parser(lang)
+        query = self.querify(query, language, translit, default_op, parser_flags)
 
-        if mlimit is None:
-            mlimit = self.backend.get_doccount()
+        if mlimit == 0:
+            mlimit = int(self.backend.get_doccount() * kmlimit)
 
-        enq.set_query(qp.parse_query(query))
+        if echo:
+            print str(query)
+        enq.set_query(query)
 
         mset = self._build_mset(enq, offset=moffset, limit=mlimit)
 
@@ -515,67 +607,29 @@ class Database(object):
             decider = PrefixDecider(prefix)
 
         if decider is None:
-            decider = LangDecider(lang)
-        eset = enq.get_eset(limit, rset, decider)
+            decider = LanguageDecider(language)
+
+        if limit == 0:
+            limit = int(self.backend.get_doccount() * klimit)
+
+        eset = enq.get_eset(limit, rset, xapian.Enquire.INCLUDE_QUERY_TERMS,
+                            1.0, decider)
 
         for item in eset.items:
             val = item[0].decode('utf8')
-            if prefix and val.startswith('X' + prefix.upper()):
-                val = "%s:%s" % (prefix, val[len(prefix) + 1:])
+            if prefix and val.startswith('X%s:' % prefix.upper()):
+                suffix = val[len(prefix) + 2:]
+                if ' ' in suffix:
+                    val = '%s:"%s"' % (prefix, suffix)
+                else:
+                    val = '%s:%s' % (prefix, suffix)
             if score:
                 yield (val, item[1])
             else:
                 yield val
 
-    def similar(self, query=None, offset=0, limit=None, order=None,
-                reverse=False, lang=None, check=None,
-                match_decider=None, match_spy=None, terms=None,
-                default_op=xapian.Query.OP_AND):
-        """ Find documents in the database most relevant to the given terms.
-
-        'query' - If *terms*=None, query is passed to xaql.suggest to
-        get a list of terms. If *terms* is set, that term list is used
-        directly and the *query* is parsed and ANDed with the elite term
-        set query.
-
-        'limit' - The number of records to return.
-
-        'lang' - The language stemmer the parser should use.
-
-        'match_decider' - A MatchDecider to apply to the query.
-
-        'terms' - Optional term list from which to derive the
-        elite set of terms to match on.
-        """
-        self.reopen()
-
-        if not terms:
-            assert query, 'Query or terms is requred'
-            suggested_terms = self.suggest(
-                query, lang=lang)
-            terms = list(suggested_terms)
-            query = None
-        elif query:
-            qp = self.prepare_query_parser(lang, default_op)
-            query = qp.parse_query(query, default_parser_flags)
-
-        enq = xapian.Enquire(self.backend)
-
-        if limit is None:
-            limit = self.backend.get_doccount()
-
-        similar_query = Query(Query.OP_ELITE_SET, terms, limit)
-        if query:
-            similar_query = Query(Query.OP_AND, [similar_query, query])
-        enq.set_query(similar_query)
-
-        mset = self._build_mset(enq, offset, limit, order, reverse,
-                                check, match_decider, match_spy)
-        return self._return_objects(mset)
-
     def _build_mset(self, enq, offset=0, limit=None, order=None, reverse=False,
                     check=None, match_decider=None, match_spy=None):
-
         if order is not None:
             if isinstance(order, basestring):
                 try:
@@ -612,12 +666,56 @@ class Database(object):
             if document:
                 yield doc
             else:
-                obj = self.unserialize(doc.get_data(),
-                                       record_factory=self.record_factory)
-                yield obj
+                data = doc.get_data()
+                typ, data = loads(data)
+                yield self.record_factory(_lookup_schema(typ).from_flat(data))
 
-    def term_counter(self, prefixes):
-        """Construct a term count spy with this instance's prefix dict."""
-        prefix_map = self.relevance_prefixes.copy()
-        prefix_map.update(self.boolean_prefixes)
-        return TermCountMatchSpy(prefixes, prefix_map)
+
+def jsonrpc_wrapper(f):
+    def w(database, *args, **kw):
+        # unstupidify the way the jsonrpc spec calls us
+        if not args:
+            args = []
+            for k in kw:
+                try:
+                    i = int(k)
+                except ValueError:
+                    continue
+                args.append((i, kw.pop(k)))
+            args.sort(key=itemgetter(0))
+            args = tuple(a[1] for a in args)
+        return f(database, *args, **kw)
+    return w
+
+
+class JSONDatabase(object):
+    """
+    Thunk layer on top of database that returns json data.
+    """
+
+    def jsonify_record(database, record):
+        return record.flatten()
+
+    def __init__(self, database):
+        database.record_factory = self.jsonify_record
+        self._database = database
+
+    @jsonrpc_wrapper
+    def query(self, *args, **kw):
+        return list(self._database.query(*args, **kw))
+
+    @jsonrpc_wrapper
+    def suggest(self, *args, **kw):
+        return list(self._database.query(*args, **kw))
+
+    @jsonrpc_wrapper
+    def facet(self, *args, **kw):
+        return self._database.facet(*args, **kw)
+
+    @jsonrpc_wrapper
+    def expand(self, *args, **kw):
+        return self._database.expand(*args, **kw)
+
+    @jsonrpc_wrapper
+    def count(self, *args, **kw):
+        return self._database.count(*args, **kw)
